/****************************************************************************/
/*      Motorola Time Processor Unit microcode assembler                    */
/****************************************************************************/
/*                                                                          */
/*      Copyright 1998, (c) by Josef Wolf <jw@raven.inka.de>                */
/*                                                                          */
/*      This file is under the GNU General Public License.                  */
/*      See LICENSE for details.                                            */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*      Example for the TPU glue code                                       */
/*                                                                          */
/****************************************************************************/

# include "tpuglue.h"
# include <stdlib.h>

/* Assuming we don't twiddle the MM bit in SIMMCR, here are the important
 * addresses:
 */
volatile unsigned short *tpubase = (unsigned short *) 0x0fffe00;
volatile unsigned short *rambase = (unsigned short *) 0x0fffb00;

/* We need a location to map the emuram while copying the uCode.
 * Make sure this value is a multiple of 0x0800
 */
# define RAMLOC 0x0e00000

int main ()
{
  int err;
  unsigned char some_byte_value;
  unsigned short some_word_value;
  volatile struct TPU_asyn_rxd_s *rcv_regs7;

  /* to initialize the TPU the following conditions must be ensured:
   * - the TPU module must be accessible
   * - the RAM must be accessible
   */
rambase[0] = 0;			/* make RAM accessible */
rambase[2] = RAMLOC >> 8;	/* at specified location */

  /* init_tpu() will
   * - copy the u-code into the emulation ram,
   * - verify the checksum,
   * - switch the TPU into emulation mode,
   * - do the requested initializations
   */
  err = init_tpu (tpubase, (void *) RAMLOC,	/* define placement */
	    TCR1CLC16, TCR2CLC8,		/* prescalings */
	    0,					/* no t2clocking-gate */
	    1);				/* permit access in user-mode */
  if (err) exit (EXIT_FAILURE);		/* There was an error */

  /* Set IRQ-arbitration, vector and level
   */
  tpu_irqc (3, 0x090, 5);

  /* Initilize a TPU channel
   */
  tpu_chan_init (7,			/* Initilize channel 7 */
		 TPU_asyn_rxd,		/* to perform the asyn_rxd function */
		 2,			/* with sequence bit 1 set */
		 TPU_HSR_INIT_RECEIVE,	/* perform the INIT_RECIEVE HSR */
		 TPU_P_HIGH,		/* Give the channel high priority */
		 1);			/* Permit IRQs for the channel */

  /* The following code is just to give you a clue how to access the
   * TPU registers. This code don't do anything meaningfull
   */

  /* this is the hard way */
  tpureg[7].asyn_rxd.CHANNEL_CONTROL.w = some_word_value;

  /* with a short-cut */
  rcv_regs7 = &tpureg[7].asyn_rxd;

  /* you can have it a little simpler */
  rcv_regs7->CHANNEL_CONTROL.w	= some_word_value;
  rcv_regs7->MATCH_RATE.b.h	= some_byte_value;
  rcv_regs7->DATA_SIZE.b.l	= some_byte_value;

  some_byte_value = rcv_regs7->RECEIVED_DATA.b.l;

  exit (EXIT_SUCCESS);
}
/* EOF */
