/****************************************************************************/
/*      Motorola Time Processor Unit microcode assembler                    */
/****************************************************************************/
/*                                                                          */
/*      Copyright 1998, (c) by Josef Wolf <jw@raven.inka.de>                */
/*                                                                          */
/*      This file is under the GNU General Public License.                  */
/*      See LICENSE for details.                                            */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*      Read in the C interface header file and intialize the TPU-stuff     */
/*                                                                          */
/****************************************************************************/

# define TPUGLUE
# include "tpuglue.h"

# define TPUMCR	0
# define TPUICR	4

volatile tpureg_t *tpureg = 0;
static volatile unsigned short *tpuloc = 0;

/* The tpuset() function gives channel based write access to specific TPU
 * registers. The regdef structure defines how those registers are to be
 * accessed.
 */
static struct regdef {
  short off, msk, bits, lshft, cmsk;
} rd[] = {
/* TPUIER */ {  5,   1,    1,     4,   15},
/* TPUFSR */ {  6,  15,    4,     2,    3},
/* TPUHSQ */ { 10,   3,    2,     3,    7},
/* TPUHSR */ { 12,   3,    2,     3,    7},
/* TPUCPR */ { 14,   3,    2,     3,    7},
};
# define TPUREG	0x080	/* offset to the channel parameter ram */
# define TPUCHMSK 0x0f

static void tpuset (int chan, int t, int value)
{
	int shft;
	struct regdef *r;
	unsigned volatile short *loc;

	if (!tpuloc) return;
	r = &rd[t];
	loc = &tpuloc[r->off + (((~chan) & TPUCHMSK) >> (r->lshft))];
	shft = (chan & r->cmsk) * r->bits;
	*loc = (*loc & ~(r->msk << shft)) | ((value & r->msk) << shft);
}

static int tpucopy (void *destination)
{
	register unsigned char *dst;
	register unsigned count;
	register unsigned long chsum = 0;

	dst = (unsigned char *) destination;
	for (count = 0; count < ((TPULOWEND + 1) * 4); count++) {
		dst[count] = tpuprog1[count];
		chsum +=     tpuprog1[count];
	}
	for (count = 0; count < ((TPUHIEND - TPUHISTART + 1) * 4); count++) {
		dst[TPUHISTART * 4 + count] = tpuprog2[count];
		chsum +=                      tpuprog2[count];
	}

	return ((chsum == TPUCHSUM) ? 0 : -1);
}

int init_tpu (	volatile void *location, void *destination,
		int tcr1p, int tcr2p, int t2cg,
		int user)
{
	unsigned short mcrval = 0;
	if (destination) {
		if (tpucopy (destination) < 0) return (-1);
		mcrval |= (1 << 10);
	}
	if (!location) return (-1);
	tpuloc = location;
	tpureg = (tpureg_t *) (&tpuloc[TPUREG]);
	mcrval |= (((tcr1p & 3) << 13)
		|  ((tcr2p & 3) << 11)
		|  ((t2cg  & 1) <<  9)
		|  ((user  & 1) <<  7)
		|  ((tcr1p & 4) <<  4)
		);
	tpuloc[TPUMCR] = mcrval;
	return (0);
}

void tpu_irqc (int iarb, int irqvec, int irqlevel)
{
	if (!tpuloc) return;
	tpuloc[TPUMCR] = (tpuloc[TPUMCR] & ~0x0f) | (iarb & 0x0f);
	tpuloc[TPUICR] = ((irqlevel & 7) << 8) | (irqvec & 0x0f0);
}

void tpu_stop ()
{
	if (!tpuloc) return;
	tpuloc[TPUMCR] |= 0x08000;
}

void tpu_start ()
{
	if (!tpuloc) return;
	tpuloc[TPUMCR] &= 0x7fff;
}

void tpu_chan_init (int chan, int func,
		int sequ, int serv,
		int prio, int irqen)
{
	if (!tpuloc) return;
	tpu_func (chan, func);
	tpu_hsq  (chan, sequ);
	tpu_hsrq (chan, serv);
	tpu_irqs (chan, irqen);
	tpu_prio (chan, prio);
}
/* EOF */
