/****************************************************************************/
/*      Motorola Time Processor Unit microcode assembler                    */
/****************************************************************************/
/*                                                                          */
/*      Copyright 1998, (c) by Josef Wolf <jw@raven.inka.de>                */
/*                                                                          */
/*      This file is under the GNU General Public License.                  */
/*      See LICENSE for details.                                            */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*      Read in the C interface header file and intialize the TPU-stuff     */
/*                                                                          */
/****************************************************************************/

# define TPUIER	0
# define TPUFSR 1
# define TPUHSQ	2
# define TPUHSR	3
# define TPUCPR 4

/* set channel 'chan' to perform function 'func'
 */
# define tpu_func(chan, func)	tpuset ((chan), TPUFSR, (func))

/* signal host service request 'serv' to channel 'chan'
 */
# define tpu_hsrq(chan, serv)	tpuset ((chan), TPUHSR, (serv))

/* set sequence number 'sequ' to channel 'chan'
 */
# define tpu_hsq(chan, sequ)    tpuset ((chan), TPUHSQ, (sequ))

/* set priority 'prio' to channel 'chan'
 */
# define tpu_prio(chan, prio)	tpuset ((chan), TPUCPR, (prio))
# define TPU_P_OFF	0
# define TPU_P_LOW	1
# define TPU_P_MIDDLE	2
# define TPU_P_HIGH	3

/* enable/disable interrupts from channel 'chan'
 */
# define tpu_irqs(chan, irqen)	tpuset ((chan), TPUIER, (irqen))

/*  initialize tpuram, tpu and switch to emulation mode
 *     location:    the place where the tpu module configuration registers
 *                  can be found
 *     destination: storage to which the tpu u-code is to be copied. This
 *                  _have_ to be the tpu emulation ram
 *     tcr1p:       rate by which the system clock is to be divided
 *     tcr2p:       rate by which the system clock is to be divided
 *     t2cg:        t2 clock gate control
 *     user:        allow user access to host seqence and host service requst
 *                  registers
 */
int init_tpu (	volatile void *location, void *destination,
		int tcr1p, int tcr2p, int t2cg,
		int user);
# define TCR1CLC4	4
# define TCR1CLC8	5
# define TCR1CLC16	6
# define TCR1CLC32	7
# define TCR1CLC64	1
# define TCR1CLC128	2
# define TCR1CLC256	3

# define TCR2CLC8	0
# define TCR2CLC16	1
# define TCR2CLC32	2
# define TCR2CLC64	3

/* set interrupt configuration
 *   iarb:     arbitration level
 *   irqvec:   interrupt vector number
 *   irqlevel: interrupt level
 */
void tpu_irqc (int iarb, int irqvec, int irqlevel);

/* place tpu ram into low power stop mode
 */
void tpu_stop ();

/* revert tpu ram from low power stop mode
 */
void tpu_start ();

/* initialize channel 'chan' to perform function 'func' with priority 'prio'
 * and sequence number 'sequ'.
 * enable/disable interrupts from and signal host service request 'serv' to it.
 */
void tpu_chan_init (int chan, int func,
		int sequ, int serv,
		int prio, int irqen);

# include "tpumcode.h"

extern volatile tpureg_t *tpureg;

/* EOF */
