/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the instruction encoding of the assembler.			*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	30/04/98	Zoltan Kocsi	Alpha version								*/
/*																			*/
/****************************************************************************/

#include "tpuasm.h"

/****************************************************************************/
/*							Global data										*/
/****************************************************************************/

INSN	insn;					/* Global instruction descriptor	*/
int		pc;						/* Current uPC						*/

/****************************************************************************/
/*					Local definitions and type								*/
/****************************************************************************/

#define	GETFLD( name, pos ) { name ## _WIDTH, &insn. ## name, pos }

typedef	struct {

	int		width;
	FIELD	*from;
	int		posit;

} ASSEMBLE;

/****************************************************************************/
/*						Local function prototypes							*/
/****************************************************************************/

static	void	Assemble( const ASSEMBLE *desc );
static	void	BusContention( int pc1, int pc2 );
static	void	TestInvalid( void );
static	void	OrganiseAU( void );

/****************************************************************************/
/*							Local data										*/
/****************************************************************************/

static	const	ASSEMBLE	format1[] = {

	GETFLD( rw,		29 ),
	GETFLD( t1abs,	25 ),
	GETFLD( t3abd,	21 ),
	GETFLD( shf,	19 ),
	GETFLD( src,	18 ),
	GETFLD( ccl,	17 ),
	GETFLD( t1bbs,	14 ),
	GETFLD( cin,	13 ),
	GETFLD( binv,	12 ),
	GETFLD( iom,	 9 ),
	GETFLD( aid,	 2 ),
	GETFLD( decend,	 0 ),
	{ 0, NULL, 30 }
};

static	const	ASSEMBLE	format2[] = {

	GETFLD( erw,	29 ),
	GETFLD( t1abs,	25 ),
	GETFLD( t3abd,	21 ),
	GETFLD( shf,	19 ),
	GETFLD( tdl,	18 ),
	GETFLD( mrl,	17 ),
	GETFLD( t1bbs,	14 ),
	GETFLD( cin,	13 ),
	GETFLD( binv,	12 ),
	GETFLD( pac,	 9 ),
	GETFLD( lsl,	 8 ),
	GETFLD( psc,	 6 ),
	GETFLD( flc,	 3 ),
	GETFLD( cir,	 2 ),
	GETFLD( decend,	 0 ),
	{ 1, NULL, 30 }
};

static	const	ASSEMBLE	format3[] = {

	GETFLD( bcc,	26 ),
	GETFLD( fls,	25 ),
	GETFLD( baf,	16 ),
	GETFLD( tbs,	12 ),
	GETFLD( pac,	 9 ),
	GETFLD( bcf,	 8 ),
	GETFLD( psc,	 6 ),
	GETFLD( flc,	 3 ),
	GETFLD( ccm,	 2 ),
	GETFLD( mtsr,	 0 ),
	{ 2, NULL,	30 }
};

static	const	ASSEMBLE	format4[] = {

	GETFLD( rw, 	28 ),
	GETFLD( nma, 	26 ),
	GETFLD( fls, 	25 ),
	GETFLD( baf, 	16 ),
	GETFLD( flc, 	13 ),
	GETFLD( lsl, 	12 ),
	GETFLD( iom, 	 9 ),
	GETFLD( aid, 	 2 ),
	GETFLD( decend,  0 ),
	{ 6, NULL,	29 }
};

static	const	ASSEMBLE	format5[] = {

	GETFLD( t1abs,	25 ),
	GETFLD( t3abd,	21 ),
	GETFLD( shf,	19 ),
	GETFLD( src,	18 ),
	GETFLD( ccl,	17 ),
	GETFLD( t1bbi,	 9 ),
	GETFLD( lsl,	 8 ),
	GETFLD( flc,	 3 ),
	GETFLD( cir,	 2 ),
	GETFLD( decend,	 0 ),
	{ 7, NULL, 29 }
};

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Initialises all fields to the default (unset) state
*/

void	InsnInitAll( void )
{
	insn.form			= FORM1 | FORM2 | FORM3 | FORM4 | FORM5;
	insn.error			= False;
	insn.rw.value		= RW_NIL;	
	insn.rmer.value		= False;
	insn.t1abs.value	= T1ABS_NIL;	
	insn.t3abd.value	= T3ABD_NIL;	
	insn.shf.value		= SHF_NIL;	
	insn.src.value		= SRC_NIL;	
	insn.ccl.value		= CCL_NIL;	
	insn.t1bbs.value	= T1BBS_NIL;	
	insn.cin.value		= CIN_NIL;	
	insn.binv.value		= BINV_NIL;	
	insn.iom.value		= IOM_NIL;	
	insn.aid.value		= AID_NIL;	
	insn.decend.value	= DECEND_NIL;	
	insn.erw.value		= ERW_NIL;	
	insn.tdl.value		= TDL_NIL;
	insn.mrl.value		= MRL_NIL;	
	insn.pac.value		= PAC_NIL;	
	insn.lsl.value		= LSL_NIL;	
	insn.psc.value		= PSC_NIL;	
	insn.flc.value		= FLC_NIL;	
	insn.cir.value		= CIR_NIL;	
	insn.bcc.value		= BCC_NIL;	
	insn.fls.value		= FLS_NIL;	
	insn.baf.value		= BAF_NIL;	
	insn.tbs.value		= TBS_NIL;	
	insn.bcf.value		= BCF_NIL;	
	insn.ccm.value		= CCM_NIL;	
	insn.mtsr.value		= MTSR_NIL;	
	insn.nma.value		= NMA_NIL;	
	insn.t1bbi.value	= T1BBI_NIL;
	insn.bcons.value	= 1;
	
	insn.rw.is_set		= False;	
	insn.rmer.is_set	= False;
	insn.t1abs.is_set	= False;	
	insn.t3abd.is_set	= False;	
	insn.shf.is_set		= False;	
	insn.src.is_set		= False;	
	insn.ccl.is_set		= False;	
	insn.t1bbs.is_set	= False;	
	insn.cin.is_set		= False;	
	insn.binv.is_set	= False;	
	insn.iom.is_set		= False;	
	insn.aid.is_set		= False;	
	insn.decend.is_set	= False;	
	insn.erw.is_set		= False;	
	insn.tdl.is_set		= False;
	insn.mrl.is_set		= False;	
	insn.pac.is_set		= False;	
	insn.lsl.is_set		= False;	
	insn.psc.is_set		= False;	
	insn.flc.is_set		= False;	
	insn.cir.is_set		= False;	
	insn.bcc.is_set		= False;	
	insn.fls.is_set		= False;	
	insn.baf.is_set		= False;	
	insn.tbs.is_set		= False;	
	insn.bcf.is_set		= False;	
	insn.ccm.is_set		= False;	
	insn.mtsr.is_set	= False;	
	insn.nma.is_set		= False;	
	insn.t1bbi.is_set	= False;
	insn.bcons.is_set	= False;
}

/*
*	Set functions for instruction fields
*/

void	SetAID( int val1, int val2 )
{
	if ( val1 < 0 || val1 > 15 || val2 < 0 || val2 > 7 ) {
	
		IssueError(( E_MEMADR ));
		val1 = AID_NIL;
		insn.error = True;
	}
	else {
		
		if ( val1 < 14 && val2 > 5 ) {
		
			IssueWarning(( E_PARMT1, val1, val2 ));
		}
		
		val1 = val1 * 8 + val2;
	}
		
	SetField( FORM1 | FORM4, &insn.aid, val1, "AID" );
}

void	SetBAF( IDENT *val )
{
	SetField( FORM3 | FORM4, &insn.baf, val->id, "BAF" );
}

void	SetBCF( int val )
{
	if ( val == 0 ) {
	
		SetField( FORM3, &insn.bcf, BCF_CLEAR, "BCF" );
	}
	else if ( val == 1 ) {
	
		SetField( FORM3, &insn.bcf, BCF_SET, "BCF" );
	}
	else {
	
		IssueError(( E_BCCVAL ));
		insn.error = True;
	}
}

void	SetNOP( void )
{
	SetField( FORM5, &insn.t1abs,  T1ABS_NIL,  "T1ABS"  );
	SetField( FORM5, &insn.t3abd,  T3ABD_NIL,  "T3ABD"  );
	SetField( FORM5, &insn.shf,    SHF_NIL,    "SHF"    );
	SetField( FORM5, &insn.src,    SRC_NIL,    "SRC"    );
	SetField( FORM5, &insn.ccl,    CCL_NIL,    "CCL"    );
	SetField( FORM5, &insn.lsl,    LSL_NIL,    "LSL"    );
	SetField( FORM5, &insn.flc,    FLC_NIL,    "FLC"    );
	SetField( FORM5, &insn.cir,    CIR_NIL,    "CIR"    );
	SetField( FORM5, &insn.t1bbi,  T1BBI_NIL,  "T1BBI"  );
	SetField( FORM5, &insn.decend, DECEND_NIL, "DECEND" );
}

void	SetRMER( void )
{
	SetField( FORM1 | FORM2 | FORM5, &insn.rmer, True, "read_mer" );
}

/*
*	This function sets the A source bus.
*	If it is set to 0 
*/

void	SetT1ABS( int val )
{
	if ( ( val == T1ABS_ZERO16 || val == T1ABS_ZERO8 ) &&
		 insn.t1abs.is_set && insn.t1abs.value == T1ABS_ZERO_MER )
	
		return;
	
	SetField( FORM1 | FORM2 | FORM5, &insn.t1abs, val, "T1ABS" );
}

/*
*	This function is a special thing, it is 
*	used so that the assembler can use any of form 1, 2 or 5 
*	with the constant
*/

void	SetBBSConstMaybe( int n )
{
	if ( n >= 2 && n <= 255 ) {
	
		SetT1BBI( n );
		return;
	}
	
	if ( n < 0 || n > 1 ) {
	
		SetBBSConst( n );
		return;
	}
	
	SetField( FORM1 | FORM2 | FORM5, &insn.bcons, n, "T1BBS or T1BBI" );
}

void	SetT1BBI( int n )
{
	if ( n < 0 || n > 255 ) {
	
		IssueError(( E_IMMDAT ));
		insn.error = True;
		n = 0;
	}

	SetField( FORM5, &insn.t1bbi, n, "T1BBI" );	
}

void	TestConstant( int data, int expect )
{
	if ( data != expect ) {
	
		IssueError(( E_CSTRNG, expect ));
		insn.error = True;
	}
}

void	SetBBSConst( int val )
{
	if ( val == -1 ) val = 0xffff;
	
	switch ( val ) {
	
		case 0:
		
			SetT1BBS( T1BBS_ZERO );
			SetBINV( BINV_PONATE );
			SetCIN( CIN_NOCARRY );
			break;
		
		case 1:
		
			SetT1BBS( T1BBS_ZERO );
			SetBINV( BINV_PONATE );
			SetCIN( CIN_CARRY );
			break;
		
		case 0xffff:
		
			SetT1BBS( T1BBS_ZERO );
			SetBINV( BINV_NEGATE );
			SetCIN( CIN_NOCARRY );
			break;
		
		case 0x8000:
		
			SetT1BBS( T1BBS_ZERO );
			SetBINV( BINV_NEGATE );
			SetCIN( CIN_CARRY );
			break;
		
		default:
		
			IssueError(( E_CSTSET ));
			insn.error = True;
			break;
	}
}

/*
*	This function sets a given field in an instruction.
*	It also checks for possible instruction format conflicts and
*	multiple settings of the same field.
*/

void	SetField( int form, FIELD *field, int val, char *name )
{
	if ( ! ( form & insn.form ) ) {
	
		IssueError(( E_INSFRM, name ));
		insn.error = True;
		return;
	}
	
	if ( field->is_set != False && val != field->value ) {
	
		IssueError(( E_FLDSET, name ));
		insn.error = True;
		return;
	}
	
	field->is_set = True;
	field->value  = val;
	insn.form	 &= form;
}

/*
*	This function is called when parsing an entire insn is finished
*/

void	InsnAssemble( void )
{
	if ( pc >= maxmem ) {
	
		IssueError(( E_MAXUPC, pc, maxmem ));
	}
	else if ( tpuram[ pc ].valid != Unknown ) {
		
		IssueError(( E_MEMFUL, pc, E_MEMTIN ));
		tpuram[ pc ].valid = False;			
	}
	else {
	
		OrganiseAU();
		tpuram[ pc ].valid = True;
		tpuram[ pc ].insn  = insn;
		tpuram[ pc ].line  = lineno;
		tpuram[ pc ].srcn  = sourcename;
		tpuram[ pc ].srcl  = src_lineno;
		cif_touch_pc (pc);
		pc++;
	}
	
	InsnInitAll();
}

/****************************************************************************/
/*							Local  functions								*/
/****************************************************************************/

/*
*	Tries to re-organise the AU subfield (if it is meaningful)
*/

static	void	OrganiseAU( void )
{
	/* If the insn has no AU field, nothing to do */

	if ( ! ( insn.form & ( FORM1 | FORM2 | FORM5 ) ) ) return;
	
	/* Check for shift / AU destination conflicts */
	
	if ( insn.src.is_set && insn.src.value == SRC_SHIFT &&
	     insn.t3abd.is_set && insn.t3abd.value == T3ABD_SR ) {
	
		IssueWarning(( E_SHRCNF ));
		insn.src.is_set = False;
		insn.src.value  = SRC_NIL;
	}
	
	/* First fix up any B constants */
		
	if ( insn.bcons.is_set ) {
		
		if ( insn.form & ( FORM1 | FORM2 ) ) {
			
			SetBBSConst( insn.bcons.value );
		}
		else {
			
			SetT1BBI( insn.bcons.value );
		}
			
		insn.bcons.is_set = False;
	}
		
	if ( insn.rmer.is_set ) {
	
		/* The au subinsn has a read_mer field. This implies that
		   the A source must be 0 (T1ABS_ZERO_MER). 
		   If the A source is zero, then it's OK, otherwise we
		   try to move A to B and set A to 0. Start thinking :-) */
		
		if ( ! insn.t1abs.is_set || 
			 insn.t1abs.value == T1ABS_ZERO16 || 
			 insn.t1abs.value == T1ABS_ZERO8 ) {
		
			/* Easy, the A source is 0 anyway, so make it ZERO_MER
			   and that's it. */
			
			insn.t1abs.is_set = True;
			insn.t1abs.value  = T1ABS_ZERO_MER;
			
			/* Fix up any pending B constants */
			
			if ( insn.bcons.is_set ) {
		
				if ( insn.form & ( FORM1 | FORM2 ) ) {
			
					SetBBSConst( insn.bcons.value );
				}
				else {
			
					SetT1BBI( insn.bcons.value );
				}
			
				insn.bcons.is_set = False;
			}
		}
		else if ( ! ( insn.form & ( FORM1 | FORM2 ) ) ) {
		
			/* The insn has no B source field that we could possibly move
			   the A source, report error */
			   
			IssueError(( E_AURDMR ));
			insn.error = True;
		}
		else if ( insn.t1abs.value != T1ABS_P && 
				  insn.t1abs.value != T1ABS_A &&
				  insn.t1abs.value != T1ABS_SR &&
				  insn.t1abs.value != T1ABS_DIOB ) {
				  
			/* The A source is something that can not be moved to be
			   B source therefore it is an error. */
			
			IssueError(( E_AURDMR ));
			insn.error = True;
		}
		else if ( insn.t1bbs.is_set && insn.t1bbs.value != T1BBS_ZERO ) {
		
			/* B is already set to a non-zero value so A can not be
			   moved to B -> report error */
		
			IssueError(( E_AURDMR ));
			insn.error = True;
		}
		else if ( insn.binv.is_set && insn.binv.value == BINV_NEGATE ) {
		
			/* The B bus was set to negation of 0, which is obviously
			   a way to add 0x8000 or 0xffff to the A source. This
			   can not be done using B bus alone, report error */
			   
			IssueError(( E_AURDMR ));
			insn.error = True;
		}
		else {
		
			/* Now we know that we can safely move A source to B. */
			
			switch ( insn.t1abs.value ) {
			
				case T1ABS_A:
				
					insn.t1bbs.value = T1BBS_A;
					break;
					
				case T1ABS_P:
				
					insn.t1bbs.value = T1BBS_P;
					break;
					
				case T1ABS_DIOB:
				
					insn.t1bbs.value = T1BBS_DIOB;
					break;
					
				case T1ABS_SR:
				
					insn.t1bbs.value = T1BBS_SR;
					break;
			}
			
			/* Do all other changes needed and we are done.*/
			
			insn.t1bbs.is_set = True;
			insn.t1abs.value  = T1ABS_ZERO_MER;
		}
	}
}

/*
*	Assembles an instruction
*
*/

static	void	Assemble( const ASSEMBLE *desc )
{
long	ins;

	ins = 0xffffffff;
	
	for ( ; desc->from ; desc++ )
	
		ins &= ( desc->from->value << desc->posit ) |
			  ~( ( ( 1 << desc->width ) - 1 ) << desc->posit );
	
	ins &= ( desc->width << desc->posit ) | 
		  ~( ( ( 1 << ( 32 - desc->posit ) ) - 1  ) << desc->posit );
	
	tpuram[ pc ].value = ins;
}

/*
*	Fixes the insns up (forward reference things)
*/

void	FixupInsn( void )
{
IDENT	*id;

	for ( pc = 0 ; pc < maxmem ; pc++ ) {
	
		if ( tpuram[ pc ].valid != True ) continue;
		insn = tpuram[ pc ].insn;
		if ( insn.error ) continue;
		
		if ( insn.baf.is_set ) {
			
			if ( ! ( id = IdentById( insn.baf.value ) ) ) {
				
				FatalError(( E_INTERN ));
			}
				
			if ( ! id->is_set ) {
				
				FixupError( tpuram[ pc ].srcn, tpuram[ pc ].srcl,
							( E_UNDSYM, id->name ));
				tpuram[ pc ].valid = False;
				continue;
			}
			else {
			
				insn.baf.value = id->value;
			}
		}
				
		if ( insn.form & FORM1 ) {
		
			Assemble( format1 );
		}
		else if ( insn.form & FORM2 ) {
		
			Assemble( format2 );
		}	
		else if ( insn.form & FORM3 ) {
		
			Assemble( format3 );
		}	
		else if ( insn.form & FORM4 ) {
		
			Assemble( format4 );
		}	
		else if ( insn.form & FORM5 ) {
		
			Assemble( format5 );
		}	
		else {
		
			FatalError(( E_INTERN ));
		}
		
		tpuram[ pc ].valid	= True;
	}
	
	TestInvalid();	
}

/*
*	This function checks for illegal insn combinations:
*
*		After writing the channel register (au chan_reg or 
*		au chan_dec) it is illegal to write the MER (chan write_mer)
*		in the next insn or read the MER (au read_mer) 
*		in the second next insn. The latter can not be checked since
*		we don't know how did we get here.
*
*		After writing to MER (chan write_mer) it is illegal to read MER
*		(au read_mer) in the next insn.
*/

static	void	TestInvalid( void )
{
	for ( pc = 0 ; pc < maxmem-1 ; pc++ ) {
	
		if ( tpuram[ pc ].valid != True ) continue;
		
		if ( tpuram[ pc ].insn.t3abd.is_set &&
			 ( tpuram[ pc ].insn.t3abd.value == T3ABD_CHAN ||
			   tpuram[ pc ].insn.t3abd.value == T3ABD_CHANDEC ) ) {
			
			
			/* Chan register is set, is it the last insn ? */
					   
			if ( tpuram[ pc ].insn.decend.is_set &&
				 tpuram[ pc ].insn.decend.value == DECEND_END ) {
		
				/* It is. Setting chan_reg was meaningles, but it causes
				   no harm. */
				   
				continue;
			}
			
			/* See if the next insn writes MER */
			
			if ( tpuram[ pc+1 ].insn.erw.is_set ) {
		
				/* It does, issue error */
				
				BusContention( pc, pc+1 );
			}
		}
		
		if ( tpuram[ pc ].insn.erw.is_set &&
			 tpuram[ pc+1 ].insn.t1abs.is_set &&
			 tpuram[ pc+1 ].insn.t1abs.value == T1ABS_ZERO_MER ) {
			 
			/* Write MER followed by read MER */
			
			BusContention( pc, pc+1 );
		}
	}
}

static	void	BusContention( int pc1, int pc2 )
{
	FixupError( tpuram[ pc2 ].srcn, tpuram[ pc2 ].srcl,
				( E_BUSCON, tpuram[ pc1 ].srcn, tpuram[ pc1 ].srcl ));
}

