/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the source list generatorof the assembler.			*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	11/05/98	Zoltan Kocsi	Alpha version								*/
/*																			*/
/****************************************************************************/

#include "tpuasm.h"

/****************************************************************************/
/*						Public data											*/
/****************************************************************************/

char	elin[ MAXLINE+1 ];				/* Buffer used by the error macros	*/

/****************************************************************************/
/*						Local prototypes									*/
/****************************************************************************/

static	void	CopyLine( FILE *list, int *cpos );
static	int		CmpFunc( int *a, int *b );

/****************************************************************************/
/*						Local data											*/
/****************************************************************************/

static	FILE	*tmp;					/* Temporary listfile				*/

/****************************************************************************/
/*						Public functions									*/
/****************************************************************************/

/*
*	This function tries to open the temporary listfile
*/

void	ListfileInit( void )
{
	if ( ! ( tmp = tmpfile() ) )
	
		FatalError(( E_FOPENE, "<temporary file>" ));
}

/*
*	This function is called by the IssueError macro
*/

void	CreateErrorEntry( void )
{
	error = True;
	fprintf( stderr, "ERROR: %s\n", elin );
	ListCurrent();
}

/*
*	This function is called by the IssueWarning macro
*/

void	CreateWarningEntry( void )
{
	fprintf( stderr, "WARNING: %s\n", elin );
	ListCurrent();
}

/*
*	This function is called by the FixupError macro
*/

void	CreateFixupError( char *src, int lin )
{
	error = True;
	fprintf( stderr, "ERROR: %s\n", elin );
	fprintf( stderr, E_LINETX, src, lin );
}

/*
*	Fatal error, immediate exit
*/

void	Fatal( void )
{
	fprintf( stderr, "FATAL ERROR: %s\n", elin );
	ListCurrent();
	exit( 1 );
}

/*
*	This function puts the next token recognised by flex to
*	the temporary listfile
*/

void	Token( const char *token, Bool macroname )
{
static	Bool	newline = True;

	/* Depending on lst_macro we either list the macro invocation itself
	   or the macro expansion */
	   
	if ( ( in_macro && ! lst_macro ) ||
		 ( macroname && lst_macro ) ) {
		 
		return;
	}
	
	if ( newline ) {
	
		fprintf( tmp, "%4d ", src_lineno );
		newline = False;
	}
	
	if ( token[ 0 ] == '\n' ) {
	
		lineno++;
		src_lineno += ! in_macro;
		newline = True;
	}
		
	fprintf( tmp, "%s", token );
}

/*
*	This function reads the temporary file back and 
*	creates the final list file.
*
*/

void	CreateListFile( char *name )
{
int		i, j, k;
Bool	first;
Bool	flag;
int		words[ MAXRAM ];
int		cpos = 0;
char	buff[ 31 ];
FILE	*list;

	if ( ! name ) return;
	
	if ( ! strcmp( name, "-" ) ) {

		list = stdout;
	}
	else {
	
		if ( ! ( list = fopen( name, "w" ) ) ) {
		
			FatalError(( E_FOPENE, name ));
		}
	}
	
	#define		N		20	/* Empiric constant ( >= 16 !) for entries :-) */
	
	/* First, set up an index array to the insns that we will list.
	   It is needed since using %org -s can cause that insns later in the
	   source to be actually earlier in the memory */
	
	for ( j = i = 0 ; i < maxmem ; i++ )
	
		if ( tpuram[ i ].valid ) words[ j++ ] = i;
		
	qsort( words, j, sizeof(int), (int(*)(const void*,const void*)) CmpFunc );
	
	rewind( tmp );
	
	for ( k = j, j = i = 0 ; i < lineno ; i++ ) {
	
		flag  = False;
		first = True;
		
		for ( ; j < k && tpuram[ words[ j ] ].line == i ; j++ ) {
			
			flag = True;
			fprintf( list, "%03X %08lX ", j, tpuram[ j ].value );
				
			if ( first ) {
				
				CopyLine( list, &cpos );
				first = False;
			}
			else
				putc( '\n', list );
		}
		
		if ( ! flag ) {
			
			fprintf( list, "             " );
			CopyLine( list, &cpos );
		}
	}
	
	fclose( tmp );
	
	/* Here we can start to generate the entry list */
	
	fprintf( list, "\n\n" );
	
	for ( i = 0 ; i < MAXENT ; i += 2 ) {
	
		if ( entries[ i ].valid == True || entries[ i+1 ].valid == True ) {
		
			fprintf( list, "%03X ", ( maxmem - 128 ) + i / 2 );
			
			if ( entries[ i ].valid == True )
			
				fprintf( list, "%04X ", entries[ i ].value );
			else
				fprintf( list, "xxxx " );
			
			if ( entries[ i+1 ].valid == True )
			
				fprintf( list, "%04X   ", entries[ i+1 ].value );
			else
				fprintf( list, "xxxx   " );
			
			if ( entries[ i ].valid == True ) {
			
				if ( entries[ i ].name ) {
				
					buff[ N ] = 0;
					strncpy( buff, entries[ i ].name, N );
					fprintf( list, buff );
					
					for ( j = N+4 - strlen( buff ) ; j > 0 ; j-- ) 
					
						putc( ' ', list );
				}
				else {
				
					fprintf( list, "<unnamed entry>" );
					for ( j = 0 ; j < N - 11 ; j++ ) putc( ' ', list );
				}
			}
			else
				for ( j = 0 ; j < N+4 ; j++ ) putc( ' ', list );
	
			if ( entries[ i+1 ].valid == True ) {
			
				if ( entries[ i+1 ].name ) {
				
					buff[ N ] = 0;
					strncpy( buff, entries[ i+1 ].name, N );
					fprintf( list, buff );
				}
				else
					fprintf( list, "<unnamed entry>" );
			}
			
			putc( '\n', list );
		}
	}	
	
	if ( list != stdout ) fclose( list );
}

/****************************************************************************/
/*						Local functions										*/
/****************************************************************************/

/*
*	This function copies one line of the tmp file to the listfile.
*	It changes tabs to spaces and truncates lines as well.
*/
	
static	void	CopyLine( FILE *list, int *cpos )
{
int		c, i, rpt;
	
	while ( ( c = getc( tmp ) ) != '\n' ) {
		
		if ( c == '\t' ) {
			
			c = ' ';
			rpt = tab_size - *cpos % tab_size;
			rpt = rpt ? rpt : tab_size;
		}
		else
			rpt = 1;
			
		for ( i = 0 ; i < rpt && *cpos < max_line ; i++, (*cpos)++ )
			
			putc( c, list );
	}
		
	putc( c, list );
	*cpos = 0;
}

			
/*
*	This function is used by quicksort to sort the line number
*	reference index array.
*/
	
static	int		CmpFunc( int *a, int *b )
{
	if ( tpuram[ *a ].line < tpuram[ *b ].line )
		
		return( -1 );
		
	else if ( tpuram[ *a ].line > tpuram[ *b ].line )
		
		return( 1 );
		
	else if ( *a < *b )
		
		return( -1 );
			
	else if ( *a > *b )
		
		return( 1 );
	else
		return( 0 ); /* This shoudl never happen ! */	
}
