%{

/********************************************************************/
/*		  Motorola Time Processor Unit microcode assembler			*/
/********************************************************************/
/*																	*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.					*/
/*																	*/
/*	This file is under the GNU General Public License.		 		*/
/*	See LICENSE for details.										*/
/*																	*/
/********************************************************************/
/*																	*/
/* This file is the parser for the assembler.						*/
/*																	*/
/********************************************************************/
/*																	*/
/*	Changelog:														*/
/*																	*/
/*	02/05/98	Zoltan Kocsi	Alpha version.						*/	
/*	19/05/98	Joern Tietjen	Fixed some syntax that Bison accep-	*/
/*								ted but yacc didn't.				*/ 
/*	26/05/98	Zoltan Kocsi	Added parser error recovery.		*/
/*																	*/
/********************************************************************/
		
#include "tpuasm.h"

static	int		p_flg;		/* P or DIOB flag ( P==1, DIOB==0) */

%}

%union {

	int		val;
	char	*string;
	IDENT	*ident;
}

	/* Keyword tokens */
	
%token	LET LET_SR LET_SL LET_RR WRITE READ
%token	AU CALL CHAN DEC_RET END GOTO IF THEN NOP RAM REPEAT RETURN
%token	A SR ERT DIOB PHIGH PLOW P LINK CHANREG DEC CHDEC TCR1 TCR2 NIL
%token	CCL SHIFT RD_MER CONST
%token	FLUSH NO_FLUSH
%token	SET CLEAR FLAG0 FLAG1
%token	PAC HIGH LOW NO_CHANGE TOGGLE LOW_HIGH HIGH_LOW NO_DETECT ANY_TRANS
%token	PIN
%token	TBS	IN_M1_C1 IN_M2_C1 IN_M1_C2 IN_M2_C2 
%token	OUT_M1_C1 OUT_M2_C1 OUT_M1_C2 OUT_M2_C2 
%token	WR_MER NEG_TDL NEG_MRL NEG_LSL CIR DIS_MTSR ENA_MTSR CONFIG
%token	LESS_THAN LOW_SAME OVF NEG CY ZERO TDL MRL LSL PSL HSQ0 HSQ1 TRUE FALSE
%token	<val> PRM
%token	BY_DIOB

	/* Other nonterminals */
	
%token	<val>		CONST 
%token	<ident>		LABEL
%token	<string>	STRING MACRO_REF

	/* Directives */
	
%token	DENTR ORG INCLUDE PAGE TYPE 
%token	NAME S_ADDR ENA_MATCH DIS_MATCH COND HSR0 HSR1 LSR MTSR X
%token	INCLUDE MACRO_DEF PAGE ORG TYPE TYPE_1 TYPE_256 TYPE_512 FUNCTION

%%

/****************************************************************************/
/*					The TPU assembly language								*/
/****************************************************************************/
	
program:		instruction
	|			program instruction
	;

instruction:	error '.' { yyerrok; }
	|			maybe_label directive
	|			maybe_label microinsn
	;
	
microinsn:		subinsn_list '.'	{ InsnAssemble(); }
	;
	
subinsn_list:	subinsn
	|			subinsn_list ';' subinsn
	;
	
maybe_label:	label maybe_label
	|
	;

label:			LABEL ':' { IdentSet( $1, pc ); }
	;
	
/****************************************************************************/
/*						Assembler directives								*/
/****************************************************************************/

directive:		entry_dir
	|			macro_dir
	|			page_dir
	|			type_dir
	|			org_dir
	|			include_dir
	;

/*
*	Format:	%page.
*	
*	Action: Sends a formfeed to the listfile
*/

page_dir:		PAGE '.'	{ DoPage(); }
	;
	

/*
*	Format:	%type type1, [512 | 256].
*
*	Action:	Sets the TPU type to 1 (there aren't any others yet) and specifies
*			the amount of TPU RAM, in 32-bit words.
*/

type_dir:		TYPE TYPE_1 ',' type_desc '.'
	;
	
type_desc:		TYPE_256	{ DoSetType( 256 ); }
	|			TYPE_512	{ DoSetType( 512 ); }
	;
	
/*
*	Format:	%org	[ address | * | label ].
*
*	Action:	Changes the PC to the given address or to the given label.
*			'*' means the current PC, that is, %org *. has no effect.
*/

org_dir:		ORG org_spec '.'
	;
	
org_spec:		LABEL { 

					if ( ! ($1)->is_set ) {
					
						IssueError(( E_ORGLAB, ($1)->name ));
					}
					else
						DoOrg( ($1)->value ); 
				}
				
	|			'*'			{ DoOrg( pc ); }
	|			CONST		{ DoOrg( $1 ); }
	;
	
/*
*	Format:	%entry { function = 0..15 | $0 .. $f ; } 
*				   { name = id , } start_address [ label | * ] ;
*				   { enable_match | disable_match ; } 
*				   cond { hsr0  = 0|1|x , }
*						{ hsr1  = 0|1|x , }
*						{ lsr   = 0|1|x , }
*						{ m/tsr = 0|1|x , }
*						{ pin   = 0|1|x , }
*						{ flag0 = 0|1|x , } ;
*				   { [ p | diob ] <- [ prm0 | .. | prm7 ] } .
*
*	Action:	Defines entry points to the TPU's event handler table.
*			When the event given by the cond mask occurs the TPU will
*			schedule execution of the function identified by the start
*			address if the user selected the function identified by the
*			function number for this channel. Enable/disable match controls
*			whether during the execution the match event latch can be asserted.
*			Name just gives a name to the function.
*			If p or diob <- xxx is specified, then when the function starts
*			the contents of the given parameter RAM location will be copied
*			to diob or p, respectively.
*/

entry_dir:		DENTR entry_opt_list '.'	{ DoEntry(); }
	;
	
entry_opt_list:	entry_opt
	|			entry_opt_list ';' entry_opt
	;
	
entry_opt:		entry_func
	|			entry_name
	|			entry_start
	|			entry_match
	|			entry_cond
	|			entry_ram
	;
	
entry_func:		FUNCTION '=' CONST	{ EntryFunc( $3 ); }
	;
	
entry_name:		NAME '=' LABEL	{ EntryName( $3 ); }
	;
	
entry_start:	S_ADDR LABEL	{ EntryStart( False, ($2)->id ); }
	|			S_ADDR '*'		{ EntryStart( True,  pc ); }
	;
	
entry_match:	ENA_MATCH		{ EntryMatch( True ); }
	|			DIS_MATCH		{ EntryMatch( False ); }
	;
	
entry_cond:		COND entry_cnd_list
	;
	
entry_cnd_list:	entry_cnd
	|			entry_cnd_list ',' entry_cnd
	;
	
entry_cnd:		HSR0  '=' CONST	{ EntryCond( EmatchHsr0,   $3 ); }
	|			HSR1  '=' CONST	{ EntryCond( EmatchHsr1,   $3 ); }
	|			LSR   '=' CONST	{ EntryCond( EmatchLsr,    $3 ); }
	|			LSR   '=' X		{ EntryCond( EmatchLsrX,   DontCare  ); }
	|			MTSR  '=' CONST	{ EntryCond( EmatchMtsr,   $3 ); }
	|			MTSR  '=' X		{ EntryCond( EmatchMtsrX,  DontCare  ); }
	|			PIN   '=' CONST	{ EntryCond( EmatchPin,    $3 ); }
	|			PIN   '=' X		{ EntryCond( EmatchPinX,   DontCare  ); }
	|			FLAG0 '=' CONST	{ EntryCond( EmatchFlag0,  $3 ); }
	|			FLAG0 '=' X		{ EntryCond( EmatchFlag0X, DontCare  ); }
	;

entry_ram:		RAM entry_reg READ entry_par
	;

entry_reg:		P				{ EntryRegp( True ); }
	|			DIOB			{ EntryRegp( False ); }
	;
	
entry_par:		PRM				{ EntryPar( $1 ); }
	;
	
/*
*	Format:	%macro name 'string'
*
*	Action:	Defines a macro. When it is dereferenced by the @name construct,
*			@name will be replaced by the contents of string.
*			(note, the dereferencing is not part of the parser, it is
*			 done in the lexing level)
*/

macro_dir:		MACRO_DEF LABEL STRING	'.'	{
				MacroDef( ($2)->name, $3 );
				cif_macro (($2)->name, $3, "");
			}
	|		MACRO_DEF LABEL STRING STRING '.' {
				MacroDef( ($2)->name, $3 );
				cif_macro (($2)->name, $3, $4);
			}
	;

/*
*	Format:	%include 'filename' { ; function = 0..15 }
*
*	Action:	The named file is read in, substituting the directive itself.
*			If function number is given, it will be the default function
*			number for the included file. 
*			Include files can be nested. If an include file has already
*			been included, a warning is issued and the second invocation
*			is ignored.
*/

include_dir:	INCLUDE STRING '.'	{ 

					DoInclude( $2 );
				}
				
	|			INCLUDE STRING ';' FUNCTION '=' CONST '.' {
	
					DoInclude( $2 );
					SetDefaultFunc( $6 );
					cif_func ($2, $6);
				}
	;
	
/****************************************************************************/
/*						Subinstructions										*/
/****************************************************************************/

subinsn:		au_sinsn 
	|			call_sinsn
	|			chan_sinsn
	|			dec_ret_sinsn
	|			end_sinsn
	|			goto_sinsn
	|			if_sinsn
	|			nop_sinsn
	|			ram_sinsn
	|			rep_sinsn
	|			ret_sinsn
	;
	

/*
*	Format:	repeat
*
*	Action:	Repeats the instruction of which repeat is a subinsns decrementer
*			times. 
*
*	Class:	1, 2, 4, 5
*
*	Clash:	end, dec_return
*/

rep_sinsn:		REPEAT		{ SetDecEnd( DECEND_REP ); }
	;

/*
*	Format:	end
*
*	Action:	Finishes the current microcode segment (state) and gives up
*			the microsequencer.
*
*	Class:	1, 2, 4, 5
*
*	Clash:	repeat, dec_return	
*/

end_sinsn:		END			{ SetDecEnd( DECEND_END ); }
	;
	
/*
*	Format:	dec_return
*
*	Action:	Starts the decrementer and when it reaches 0 then uPC 
*			will be reload from the ret. addr. reg (that is, it is
*			a timed return from a subroutine - each insn takes one
*			decrementer tick to execute).
*
*	Class:	1, 2, 4, 5
*
*	Clash:	repeat, end
*/

						
dec_ret_sinsn:	DEC_RET		{ SetDecEnd( DECEND_RET ); }
	;

/*
*	Format:	return {, flush | no_flush }
*
*	Action:	uPC is reload from the return address register (RAR). The flush 
*			option determines whether the insn after the one containing
*			the return will be executed before returning (it is already
*			in the prefetch queue) or will be flushed from the prefetch
*			and substituted by a NOP (it still takes time !).
*
*	Class:	4
*
*	Clash:	goto, call
*/

ret_sinsn:	RETURN {

				SetNMA( NMA_RET );
				SetFLS( FLS_NIL ); 
			}
			
	|		RETURN ',' FLUSH { 
	
				SetNMA( NMA_RET );
				SetFLS( FLS_FLUSH ); 
			}
			
	|		RETURN ',' NO_FLUSH { 
	
				SetNMA( NMA_RET );
				SetFLS( FLS_NIL ); 
			}
	;
	

/*
*	Format:	goto label {, flush | no_flush }
*
*	Action:	uPC is load to label. The flush as in return.
*
*	Class:	4
*
*	Clash:	return, call
*/

goto_sinsn:	GOTO LABEL { 

				SetBAF( $2 );
				SetNMA( NMA_GOTO );
				SetFLS( FLS_NIL ); 
			}
			
	|		GOTO LABEL ',' FLUSH {
	
				SetBAF( $2 );
				SetNMA( NMA_GOTO );
				SetFLS( FLS_FLUSH ); 
			}
			
	|		GOTO LABEL ',' NO_FLUSH { 
	
				SetBAF( $2 );
				SetNMA( NMA_CALL );
				SetFLS( FLS_NIL ); 
			}
	;
	
/*
*	Format:	call label {, flush | no_flush }
*
*	Action:	uPC is load to label. If flush, then RAR is loaded
*			with current uPC + 1. If not flush, RAR <- uPC + 2;
*			Other effects of flush as in return.
*
*	Class:	4
*
*	Clash:	return, call
*/

call_sinsn:	CALL LABEL { 

				SetBAF( $2 );
				SetNMA( NMA_CALL );
				SetFLS( FLS_NIL ); 
			}
			
	|		CALL LABEL ',' FLUSH {
	
				SetBAF( $2 );
				SetNMA( NMA_CALL );
				SetFLS( FLS_FLUSH ); 
			}
			
	|		CALL LABEL ',' NO_FLUSH { 
	
				SetBAF( $2 );
				SetNMA( NMA_CALL );
				SetFLS( FLS_NIL ); 
			}
	;
	
/*
*	Format:	if {<cond> =} <condval> then goto label {, flush | noflush }
*			Where cons is less_than, low_same, v, n, c, z, flag0, flag1,
*			tdl, mrl, lsr, hsq0, hsq1, psl. For their meaning see the lex
*			file. Condval is 1 or 0 or true ( == 1 ) or false ( == 0 ).
*
*	Action:	Executes a jump if the given condition is equal to that is
*			given in condval, otherwise continues execution on the next
*			insn.
*			If cond is not given, then condval describes an unconditional
*			jump (or no jump). This is good because that way you can make
*			a goto in a Class 3 insn.
*			The flush stuff is the same as with goto.
*
*	Class:	3
*
*	Clash:	None
*/

if_sinsn:	IF CONST THEN GOTO LABEL {

				SetBAF( $5 );
				SetFLS( FLS_NIL );
				SetBCC( BCC_NEVER );
				SetBCF( 1-$2 );
			}
			
	|		IF CONST THEN GOTO LABEL ',' FLUSH {

				SetBAF( $5 );
				SetFLS( FLS_FLUSH );
				SetBCC( BCC_NEVER );
				SetBCF( 1-$2 );
			}

	|		IF CONST THEN GOTO LABEL ',' NO_FLUSH {

				SetBAF( $5 );
				SetFLS( FLS_NIL );
				SetBCC( BCC_NEVER );
				SetBCF( 1-$2 );
			}

	|		IF condition '=' CONST THEN GOTO LABEL {

				SetBAF( $7 );
				SetFLS( FLS_NIL );
				SetBCF( $4 );
			}

	|		IF condition '=' CONST THEN GOTO LABEL ',' FLUSH {

				SetBAF( $7 );
				SetFLS( FLS_FLUSH );
				SetBCF( $4 );
			}

	|		IF condition '=' CONST THEN GOTO LABEL ',' NO_FLUSH {

				SetBAF( $7 );
				SetFLS( FLS_NIL );
				SetBCF( $4 );
			}
	;

condition:	LOW_SAME	{ SetBCC( BCC_LOWSAME );	}
	|		LESS_THAN	{ SetBCC( BCC_LESSTHAN );	}
	|		NEG			{ SetBCC( BCC_NEG );		}
	|		CY			{ SetBCC( BCC_CY );			}
	|		ZERO		{ SetBCC( BCC_ZERO );		}
	|		OVF			{ SetBCC( BCC_OVF ); 		}
	|		PSL			{ SetBCC( BCC_PSL ); 		}
	|		LSL			{ SetBCC( BCC_LSL ); 		}
	|		TDL			{ SetBCC( BCC_TDL ); 		}
	|		MRL			{ SetBCC( BCC_MRL ); 		}
	|		FLAG0		{ SetBCC( BCC_FLAG0 ); 		}
	|		FLAG1		{ SetBCC( BCC_FLAG1 ); 		}
	|		HSQ0		{ SetBCC( BCC_HSQ0 ); 		}
	|		HSQ1		{ SetBCC( BCC_HSQ1 ); 		}
	;

/*
*	Format:	nop
*
*	Action:	none.
*
*	Class:	Actually uses class 5 (the code is 0xffffffff) but it should 
*			not matter
*
*	Clash:	With everything. Nop supposed to be alone as in label: nop.
*			Its sole purpose is wasting one unit of time (and thus allow
*			pipelined HW operations to complete).
*/

nop_sinsn:	NOP		{ SetNOP(); }
	;

/*
*	Format:	ram <ram_reg> <dir> <address>
*			where ram_reg is either p or diob, dir is <- or -> and
*			address defines the RAM location as follows:
*			prm0 - prm7 identifies the channel's parameter registers,
*			prm 6 and 7 being available only on channels ???
*			by_diob means indirect address, the RAM is addressed by
*			the contents of diob. The format (C,P) where C and P are
*			constants (0-15,0-7) accesses the C-th channel's P-th
*			parameter (that is, the full parameter RAM can be accessed this
*			way since parameter locations are continuous).
*
*	Action:	Reads or writes a ram location
*
*	Class:	1 and 4
*
*	Clash:	none
*/

ram_sinsn:	RAM ram_reg read_write ram_loc
	|		ram_reg read_write ram_loc
	;
	
ram_reg:	P 		{ p_flg = 1; }
	|		DIOB	{ p_flg = 0; }
	;
	
read_write:	READ 	{ SetRW( RW_READ  ); }
	|		WRITE	{ SetRW( RW_WRITE ); }
	;
	

ram_loc:	PRM { 

				SetIOM( p_flg ? IOM_P_PRM : IOM_DIOB_PRM ); 
				SetAID( 0, $1 );
			}
			
	|		BY_DIOB {

				SetAID( 0, 0 );
				SetIOM( p_flg ? IOM_P_DIOB : IOM_DIOB_DIOB );
			}
	
	|		'(' CONST ',' CONST ')'	{ 
	
				SetIOM( p_flg ? IOM_P_AID : IOM_DIOB_AID ); 
				SetAID( $2, $4 ); 
			}
			
	|		CONST {
	
				SetIOM( p_flg ? IOM_P_AID : IOM_DIOB_AID ); 
				SetAID( $1 / 16, ( $1 / 2 ) % 8 ); 
			}
	;
	
/*
*	Format:	chan <all sorts of different operands separated by commas>
*
*	Action:	Depends on the operands. Not all operand combinations are allowed,
*			here is how it works (paren shows the insn class):
*
*		Sets or clears a user flag:
*
*	(2345)	[ set | clear ] [ flag0 | flag1 ]	
*
*		Controls the pin action. In the first 4 cases the pin is output and
*		the action on a match event is as given. The second 4 cases are 
*		when the pin is input and the given transition is detected.
*
*	(23)	pac := [ high | low | no_change | toggle | low_high | high_low
*					 no_detect | any_trans ]	: Pin control
*
*		Controls the pin and timebases for match and capture.
*
*	(3)		tbs := [in|out]_m[1|2]_c[1|2]
*
*		Pin value control, forces the pin to be 0 or 1 (or PAC)
*
*	(23)	pin := [ high | low | pac ]	
*
*		Negate the link service latch
*
*	(245)	neg_lsl
*
*		Negate transition detect latch
*
*	(2)		neg_tdl
*
*		Negate match recognition latch
*
*	(2)		neg_mrl
*
*		Request for a channel interrupt (TPU IT to the CPU)
*
*	(25)	cir
*
*		Configuration control: { TBS, PAC, PSC } = P
*
*	(3)		config := p
*
*		Service request enable/disable
*
*	(3)		[ enable_mtsr | disable_mtsr ]
*
*		Write match event register
*
*	(2)		write_mer
*
*	Class:	2, 3, 4, 5
*
*	Clash:	With itself :-)
*/

chan_sinsn:	CHAN chan_ctrl_list
	|		chan_ctrl_list
	;
	
chan_ctrl_list:	chan_ctrl
	|			chan_ctrl_list ',' chan_ctrl
	;
	
chan_ctrl:		flag_ctrl
	|			pac_ctrl
	|			pin_ctrl
	|			mer_ctrl
	|			tdl_ctrl
	|			mrl_ctrl
	|			lsl_ctrl
	|			cir_ctrl
	|			tbs_ctrl
	|			conf_ctrl
	|			mtsr_ctrl
	;
	
flag_ctrl:		SET FLAG0			{ SetFLC( FLC_SET_F0 ); }
	|			SET FLAG1			{ SetFLC( FLC_SET_F1 ); }
	|			CLEAR FLAG0			{ SetFLC( FLC_CLR_F0 ); }
	|			CLEAR FLAG1			{ SetFLC( FLC_CLR_F1 ); }
	;
	

pac_ctrl:		PAC LET HIGH		{ SetPAC( PAC_HIGH );		}
	|			PAC LET LOW			{ SetPAC( PAC_LOW );		}
	|			PAC LET NO_CHANGE	{ SetPAC( PAC_NO_CHANGE );	}
	|			PAC LET TOGGLE		{ SetPAC( PAC_TOGGLE );		}
	|			PAC LET LOW_HIGH	{ SetPAC( PAC_HIGH );		}
	|			PAC LET HIGH_LOW	{ SetPAC( PAC_LOW );		}
	|			PAC LET NO_DETECT	{ SetPAC( PAC_NO_CHANGE );	}
	|			PAC LET ANY_TRANS	{ SetPAC( PAC_TOGGLE );		}
	;
	
pin_ctrl:		PIN LET HIGH		{ SetPSC( PSC_HIGH );	}
	|			PIN LET LOW			{ SetPSC( PSC_LOW );	}
	|			PIN LET PAC			{ SetPSC( PSC_PAC );	}
	;
	
mer_ctrl:		WR_MER	{ SetERW(); }
	;
	
tdl_ctrl:		NEG_TDL { SetTDL(); }
	;

mrl_ctrl:		NEG_MRL { SetMRL(); }
	;

lsl_ctrl:		NEG_LSL { SetLSL(); }
	;
	
cir_ctrl:		CIR		{ SetCIR(); }
	;

tbs_ctrl:		TBS LET IN_M1_C1	{ SetTBS( TBS_IN11 );	}
	|			TBS LET IN_M1_C2	{ SetTBS( TBS_IN21 );	}
	|			TBS LET IN_M2_C1	{ SetTBS( TBS_IN12 );	}
	|			TBS LET IN_M2_C2	{ SetTBS( TBS_IN22 );	}
	|			TBS LET OUT_M1_C1	{ SetTBS( TBS_OUT11 ); 	}
	|			TBS LET OUT_M1_C2	{ SetTBS( TBS_OUT21 );	}
	|			TBS LET OUT_M2_C1	{ SetTBS( TBS_OUT12 );	}
	|			TBS LET OUT_M2_C2	{ SetTBS( TBS_OUT22 );	}
	;
	

conf_ctrl:		CONFIG LET P 	{ SetCCM(); }
	;
	
mtsr_ctrl:		ENA_MTSR	{ SetMTSR( MTSR_ENA ); }
	|			DIS_MTSR	{ SetMTSR( MTSR_DIS ); }
	;

/*
*	Format:	au <dest> <op> [ <const> | <expr> ] {, ccl} {, shift} {,read_mer}
*	
*	Action: There's just too much to describe ...
*
*	Class:	1, 2, 5
*
*	Clash:	Nothing, really
*/

au_sinsn:		AU au_option_list
	|			au_option_list
	;
		
au_option_list:	au_option_list ',' au_option
	|			au_option
	;
	
au_option:		au_expr
	|			CCL			{ SetCCL( CCL_LATCH ); }
	|			SHIFT		{ SetSRC( SRC_SHIFT ); }
	|			RD_MER		{ SetRMER(); }
	;

au_expr:		au_dest au_operator au_opexp
	;
	
au_dest:		A			{ SetT3ABD( T3ABD_A ); }
	|			SR			{ SetT3ABD( T3ABD_SR ); }
	|			ERT			{ SetT3ABD( T3ABD_ERT ); }
	|			DIOB		{ SetT3ABD( T3ABD_DIOB ); }
	|			PHIGH		{ SetT3ABD( T3ABD_P_HIGH ); }
	|			PLOW		{ SetT3ABD( T3ABD_P_LOW ); }
	|			P			{ SetT3ABD( T3ABD_P ); }
	|			LINK		{ SetT3ABD( T3ABD_LINK ); }
	|			CHANREG		{ SetT3ABD( T3ABD_CHAN ); }
	|			DEC			{ SetT3ABD( T3ABD_DEC ); }
	|			CHDEC		{ SetT3ABD( T3ABD_CHANDEC ); }
	|			TCR1		{ SetT3ABD( T3ABD_TCR1 ); }
	|			TCR2		{ SetT3ABD( T3ABD_TCR2 ); }
	|			NIL			{ SetT3ABD( T3ABD_NIL ); }
	;
	
au_operator:	LET			{ SetSHF( SHF_NOSHIFT ); }
	|			LET_SR		{ SetSHF( SHF_RIGHT ); }
	|			LET_SL		{ SetSHF( SHF_LEFT ); }
	|			LET_RR		{ SetSHF( SHF_ROTR ); }
	;
	
au_opexp:		CONST { 

					SetT1ABS( T1ABS_ZERO16 );
					SetBBSConstMaybe( $1 ); 
				}
				
	|			'!' CONST { 

					SetT1ABS( T1ABS_ZERO16 );
					SetBBSConstMaybe( $2 ^ 0xffff ); 
				}
				
	|			au_asrc	{

					SetBBSConstMaybe( 0 ); 
				}
	
	|			au_asrc '+' CONST {
	
					SetBBSConstMaybe( $3 );
				}
	
	|			au_asrc '-' CONST {
	
					TestConstant( $3, 1 );
					SetBBSConst( 0xffff );
				}
				
	|			au_asrc '+' au_bsrc {
	
					SetBINV( BINV_PONATE );
					SetCIN( CIN_NOCARRY );
				}
				
	|			au_asrc '+' au_bsrc '+' CONST {
	
					TestConstant( $5, 1 );
					SetBINV( BINV_PONATE );
					SetCIN( CIN_CARRY );
				}
				
	|			au_asrc '-' au_bsrc {
	
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_CARRY );
				}
					
	|			au_asrc '-' au_bsrc '-' CONST {
	
					TestConstant( $5, 1 );
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_NOCARRY );
				}
					
					
	|			au_asrc '+' '!' au_bsrc {
		
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_NOCARRY );
				}
	
	|			au_asrc '+' '!' au_bsrc '+' CONST {
	
					TestConstant( $6, 1 );
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_CARRY );
				}
				
	|			'!' au_bsrc {
		
					SetT1ABS( T1ABS_ZERO16 );
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_NOCARRY );
				}
	
	|			'!' au_bsrc '+' CONST {
	
					SetT1ABS( T1ABS_ZERO16 );
					TestConstant( $4, 1 );
					SetBINV( BINV_NEGATE );
					SetCIN( CIN_CARRY );
				}
				
	|			'#' CONST {
	
					SetT1ABS( T1ABS_ZERO16 );
					SetT1BBI( $2 );
				}
				
	|			au_asrc '+' '#' CONST {
	
					SetT1BBI( $4 );
				}
	;
	

au_asrc:		P			{ SetT1ABS( T1ABS_P ); }
	|			A			{ SetT1ABS( T1ABS_A ); }
	|			SR			{ SetT1ABS( T1ABS_SR ); }
	|			DIOB		{ SetT1ABS( T1ABS_DIOB ); }
	|			TCR1		{ SetT1ABS( T1ABS_TCR1 ); }
	|			TCR2		{ SetT1ABS( T1ABS_TCR2 ); }
	|			ERT			{ SetT1ABS( T1ABS_ERT ); }
	|			PLOW		{ SetT1ABS( T1ABS_P_LOW ); }
	|			PHIGH		{ SetT1ABS( T1ABS_P_HIGH ); }
	|			DEC			{ SetT1ABS( T1ABS_DEC ); }
	|			CHANREG		{ SetT1ABS( T1ABS_CHAN ); }
	;
	
au_bsrc:		P			{ SetT1BBS( T1BBS_P ); }
	|			A			{ SetT1BBS( T1BBS_A ); }
	|			SR			{ SetT1BBS( T1BBS_SR ); }
	|			DIOB		{ SetT1BBS( T1BBS_DIOB ); }
	;
	

%%

void yyerror( char *s )
{
	IssueError(( E_PARSER, s ));
}
