/****************************************************************************/
/*			  Motorola Time Processor Unit microcode assembler 				*/
/****************************************************************************/
/*																			*/
/*	Copyright 1998, (c) Bendor Research Pty. Ltd.							*/
/*																			*/
/*	This file is under the GNU General Public License.						*/
/*	See LICENSE for details.												*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	This file contains the S record generator of the assembler.				*/
/*																			*/
/****************************************************************************/
/*																			*/
/*	Changelog:																*/
/*																			*/
/*	11/05/98	Zoltan Kocsi	Alpha version								*/
/*																			*/
/****************************************************************************/

#include "tpuasm.h"

/****************************************************************************/
/*							Local definitions								*/
/****************************************************************************/

#define	MAXSREC	500

/****************************************************************************/
/*							Local data										*/
/****************************************************************************/

static	char	line[ MAXSREC ];
static	uchar	csum;
static	ulong	nexta;
static	ulong	start;
static	int		bcnt;
static	FILE	*fp;
static	int		srec_type;
static	int		max_bytes;

/****************************************************************************/
/*							Local prototypes								*/
/****************************************************************************/

static	void	SrecByte( uchar data, char *line );
static	void	SrecFlush( void );

/****************************************************************************/
/*							Public functions								*/
/****************************************************************************/

/*
*	Initialises the SREC file. 
*/

void	SrecOpen( char *srecname, int type, int maxbytes )
{
	if ( ! srecname ) return;
	
	if ( ! strcmp( srecname, "-" ) ) {
	
		fp = stdout;
	}
	else {
	
		if ( ! ( fp = fopen( srecname, "w" ) ) ) {
	
			FatalError(( E_FOPENE, srecname ));
		}
	}
	
	srec_type = type;
	max_bytes = maxbytes;
	
	line[0] = 0;
	csum	= 0;
	bcnt	= 0;
	nexta	= 0;
	start	= 0;
}

/*
*	Sets the address for the next data item
*/

void	SrecAddress( ulong addr )
{
	if ( addr != nexta ) {
	
		SrecFlush();
		nexta = addr;
		start = addr;
	}
}

/*
*	Writes a 4 byte data entity
*/

void	SrecLong( ulong data )
{
	SrecByte( ( data >> 24 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
	SrecByte( ( data >> 16 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
	SrecByte( ( data >>  8 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
	SrecByte( ( data >>  0 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
}

/*
*	Writes a 2 byte data entity
*/

void	SrecWord( ushort data )
{
	SrecByte( ( data >>  8 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
	SrecByte( ( data >>  0 ) & 0xff, line );	nexta++;
	if ( ++bcnt == max_bytes ) SrecFlush();
}

/*
*	Finishes the SREC file and closes it
*/

void	SrecClose( void )
{
	SrecFlush();
	fprintf( fp, "S9030000FC\n" );
	if ( fp != stdout ) fclose( fp );	
}

/****************************************************************************/
/*							Local functions									*/
/****************************************************************************/

/*
*	Writes a single byte to the given buffer and 
*	updates the checksum.
*/

static	void	SrecByte( uchar data, char *line )
{
char	buff[ 3 ];

	csum += data;
	sprintf( buff, "%02X", data );
	strcat( line, buff );
}

/*
*	Flushes the current srecord buffer and initialises for the
*	next one.
*/

static	void	SrecFlush( void )
{
int		st;
char	alin[ 9 ];

	if ( ! bcnt ) return;
	
	alin[0] = 0;
	
	/* Work out the startess length first */
	
	st = 1;
	
	if ( srec_type == 3 || start > 0x00ffffff ) {
	
		SrecByte( ( start >> 24 ) & 0xff, alin );
		st = 3;
	}
	
	if ( srec_type >= 2 || start > 0x0000ffff ) {
	
		SrecByte( ( start >> 16 ) & 0xff, alin );
		if ( st == 1 ) st = 2;
	}
	
	SrecByte( ( start >> 8 ) & 0xff, alin );
	SrecByte( start & 0xff, alin );
	bcnt += 2 + st;
	csum += bcnt;
	fprintf( fp, "S%d%02X%s%s%02X\n", st, bcnt, alin, line, ~csum & 0xff );
	line[0] = 0;
	bcnt    = 0;
	csum    = 0;
	start   = nexta;
}
